# :coding: utf-8

import os
import sys
import tempfile
import textwrap

import click
import qip
import qip.logging
import wiz
import wiz.config
from qip import __version__

# Initiate logging handler to display potential warning when fetching config.
wiz.logging.configure()

#: Retrieve configuration mapping to initialize default values.
_CONFIG = wiz.config.fetch()

#: Click default context for all commands.
CONTEXT_SETTINGS = dict(
    max_content_width=90,
    help_option_names=["-h", "--help"],
)


@click.group(
    context_settings=CONTEXT_SETTINGS
)
@click.version_option(version=__version__)
@click.option(
    "-v", "--verbosity",
    help="Set the logging output verbosity.",
    type=click.Choice(qip.logging.levels),
    default="info"
)
def main(**kwargs):
    """Quarantine Installer for Python."""
    qip.logging.configure()
    qip.logging.root.handlers["stderr"].filterer.min = kwargs["verbosity"]


@main.command(
    name="install",
    help=textwrap.dedent(
        """
        Install one or several packages.
    
        Command example:

        \b
        >>> qip install .
        >>> qip install /path/to/foo/
        >>> qip install foo
        >>> qip install foo bar
        >>> qip install "foo==0.1.0"
        >>> qip install "foo >= 7, < 8"
        >>> qip install "git@gitlab:rnd/foo.git"
        >>> qip install "git@gitlab:rnd/foo.git@0.1.0"
        >>> qip install "git@gitlab:rnd/foo.git@dev"
        >>> qip install foo -p "python==3.6.*"
        """
    ),
    short_help="Install one or several packages.",
    context_settings=CONTEXT_SETTINGS
)
@click.option(
    "-o", "--output-path",
    help=(
        "Destination for the package installation data. Default is "
        "'<TEMPORARY_FOLDER>/qip/packages'"
    ),
    type=click.Path(),
    metavar="PATH",
    default=(
        _CONFIG.get("qip", {}).get(
            "packages_output", os.path.join(
                tempfile.gettempdir(), "qip", "packages"
            )
        )
    ),
)
@click.option(
    "-d", "--definition-path",
    help=(
        "Destination for the Wiz definitions extracted. Default is "
        "'<TEMPORARY_FOLDER>/qip/definitions'"
    ),
    type=click.Path(),
    metavar="PATH",
    default=(
        _CONFIG.get("qip", {}).get(
            "definitions_output", os.path.join(
                tempfile.gettempdir(), "qip", "definitions"
            )
        )
    ),
)
@click.option(
    "-u", "--update",
    help=(
        "Update Wiz definition(s) that already exist in the Wiz definitions "
        "output path with additional Python variants."
    ),
    is_flag=True,
    default=False
)
@click.option(
    "-N", "--no-dependencies",
    help="Install the specified package(s) without dependencies.",
    is_flag=True,
    default=False
)
@click.option(
    "-f/-s", "--overwrite-installed/--skip-installed",
    help=(
        "Indicate whether packages already installed should be overwritten "
        "or skipped. By default, a user confirmation will be required."
    ),
    default=None
)
@click.option(
    "-e", "--editable",
    help=(
        "Install a project in editable mode (i.e. setuptools \"develop mode\") "
        "from a local project path or a VCS url."
    ),
    is_flag=True,
    default=False
)
@click.option(
    "-p", "--python",
    help=(
        "Target a specific Python version via a Wiz request or a path to a "
        "Python executable."
    ),
    show_default=True,
    metavar="TARGET",
    default=_CONFIG.get("qip", {}).get("python_target", sys.executable)
)
@click.argument(
    "requests",
    nargs=-1,
    required=True
)
def install(**kwargs):
    """Install one or several packages."""
    logger = qip.logging.Logger(__name__ + ".install")

    output_path = kwargs["output_path"]
    definition_path = kwargs["definition_path"]

    # Fetch definition mapping from definition path if previously extracted
    # definitions should to create new definition.
    definition_mapping = None

    if kwargs["update"]:
        definition_mapping = wiz.fetch_definition_mapping([definition_path])

    qip.install(
        kwargs["requests"], output_path,
        definition_path=definition_path,
        overwrite=kwargs["overwrite_installed"],
        no_dependencies=kwargs["no_dependencies"],
        editable_mode=kwargs["editable"],
        python_target=kwargs["python"],
        definition_mapping=definition_mapping
    )

    logger.info("Package output directory: {!r}".format(output_path))
    logger.info("Definition output directory: {!r}".format(definition_path))



# :coding: utf-8

import os
import sys
import shutil
import tempfile

import click
import pytest
import wiz.filesystem

import qip
import qip.package
import qip.definition


@pytest.fixture()
def mocked_tempfile_mkdtemp(mocker):
    """Return mocked 'tempfile.mkdtemp' function"""
    return mocker.patch.object(tempfile, "mkdtemp")


@pytest.fixture()
def mocked_shutil_rmtree(mocker):
    """Return mocked 'shutil.rmtree' function"""
    return mocker.patch.object(shutil, "rmtree")


@pytest.fixture()
def mocked_shutil_copytree(mocker):
    """Return mocked 'shutil.copytree' function"""
    return mocker.patch.object(shutil, "copytree")


@pytest.fixture()
def mocked_click_prompt(mocker):
    """Return mocked 'click.prompt' function"""
    return mocker.patch.object(click, "prompt")


@pytest.fixture()
def mocked_confirm_overwrite(mocker):
    """Return mocked 'qip._confirm_overwrite' function"""
    return mocker.patch.object(qip, "_confirm_overwrite")


@pytest.fixture()
def mocked_copy_to_destination(mocker):
    """Return mocked 'qip.copy_to_destination' function"""
    return mocker.patch.object(qip, "copy_to_destination")


@pytest.fixture()
def mocked_fetch_context_mapping(mocker):
    """Return mocked 'qip.fetch_context_mapping' function"""
    return mocker.patch.object(qip, "fetch_context_mapping")


@pytest.fixture()
def mocked_filesystem_ensure_directory(mocker):
    """Return mocked 'wiz.filesystem.ensure_directory' function"""
    return mocker.patch.object(wiz.filesystem, "ensure_directory")


@pytest.fixture()
def mocked_package_install(mocker):
    """Return mocked 'qip.package.install' function"""
    return mocker.patch.object(qip.package, "install")


@pytest.fixture()
def mocked_definition_export(mocker):
    """Return mocked 'qip.definition.export' function"""
    return mocker.patch.object(qip.definition, "export")


@pytest.fixture()
def mocked_fetch_environ(mocker):
    """Return mocked 'qip.environ.fetch' function"""
    return mocker.patch.object(qip.environ, "fetch")


@pytest.fixture()
def mocked_fetch_python_mapping(mocker):
    """Return mocked 'qip.environ.fetch_python_mapping' function"""
    return mocker.patch.object(qip.environ, "fetch_python_mapping")


@pytest.mark.parametrize(
    "options, overwrite, editable_mode, python_target", [
        ({}, False, False, sys.executable),
        ({"overwrite": True}, True, False, sys.executable),
        ({"editable_mode": True}, False, True, sys.executable),
        ({"python_target": "/bin/python3"}, False, False, "/bin/python3")
    ], ids=[
        "no-options",
        "with-overwrite-packages",
        "with-editable-mode",
        "with-python-target",
    ]
)
def test_install(
    mocked_filesystem_ensure_directory, mocked_tempfile_mkdtemp,
    mocked_fetch_context_mapping, mocked_package_install,
    mocked_copy_to_destination, mocked_definition_export,
    mocked_shutil_rmtree, options, overwrite, editable_mode, python_target
):
    """Install packages."""
    packages = [
        {
            "identifier": "Foo-0.2.3",
            "name": "Foo",
            "version": "0.2.3",
            "requirements": [
                "bim >= 3, < 4",
                "bar",
            ]
        },
        {
            "identifier": "Bar-22.3",
            "name": "Bar",
            "version": "22.3",
            "requirements": [
                "foo",
            ]
        },
        {
            "identifier": "Bim-3.2.1",
            "name": "Bim",
            "version": "3.2.1",
            "requirements": [
                "bar > 22",
            ]
        },
        {
            "identifier": "Bar-22.3",
            "name": "Bar",
            "version": "22.3",
            "requirements": [
                "foo",
            ]
        }
    ]

    context = {
        "environ": {
            "PYTHONPATH": "/path/to/site-packages"
        }
    }

    mocked_tempfile_mkdtemp.side_effect = ["/tmp1", "/tmp2"]
    mocked_fetch_context_mapping.return_value = context
    mocked_package_install.side_effect = packages

    mocked_copy_to_destination.side_effect = [
        (True, overwrite), (True, overwrite), (True, overwrite)
    ]

    qip.install(["foo", "bar"], "/path/to/install", **options)

    assert mocked_tempfile_mkdtemp.call_count == 2

    assert mocked_filesystem_ensure_directory.call_count == 9
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/install")
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/site-packages")
    mocked_filesystem_ensure_directory.assert_any_call("/tmp2")

    mocked_fetch_context_mapping.assert_called_once_with("/tmp2", python_target)

    assert mocked_package_install.call_count == 4
    mocked_package_install.assert_any_call(
        "foo", "/tmp2", context, "/tmp1",
        editable_mode=editable_mode
    )
    mocked_package_install.assert_any_call(
        "bar", "/tmp2", context, "/tmp1",
        editable_mode=False
    )
    mocked_package_install.assert_any_call(
        "bim >= 3, < 4", "/tmp2", context, "/tmp1",
        editable_mode=False
    )
    mocked_package_install.assert_any_call(
        "bar > 22", "/tmp2", context, "/tmp1",
        editable_mode=False
    )

    assert mocked_copy_to_destination.call_count == 3
    mocked_copy_to_destination.assert_any_call(
        packages[0], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )
    mocked_copy_to_destination.assert_any_call(
        packages[1], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )
    mocked_copy_to_destination.assert_any_call(
        packages[2], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )

    mocked_definition_export.assert_not_called()

    assert mocked_shutil_rmtree.call_count == 6
    mocked_shutil_rmtree.assert_any_call("/tmp1")
    mocked_shutil_rmtree.assert_any_call("/tmp2")


@pytest.mark.parametrize(
    "options, overwrite, editable_mode, python_target, definition_mapping", [
        ({}, False, False, sys.executable, None),
        ({"overwrite": True}, True, False, sys.executable, None),
        ({"editable_mode": True}, False, True, sys.executable, None),
        ({"python_target": "/bin/python3"}, False, False, "/bin/python3", None),
        (
            {"definition_mapping": "__MAPPING__"},
            False, False, sys.executable, "__MAPPING__"
        ),
    ], ids=[
        "no-options",
        "with-overwrite-packages",
        "with-editable-mode",
        "with-python-target",
        "with-definition-mapping",
    ]
)
def test_install_with_definition_path(
    mocked_filesystem_ensure_directory, mocked_tempfile_mkdtemp,
    mocked_fetch_context_mapping, mocked_package_install,
    mocked_copy_to_destination, mocked_definition_export,
    mocked_shutil_rmtree, options, overwrite, editable_mode, python_target,
    definition_mapping
):
    """Install packages with Wiz definitions."""
    packages = [
        {
            "identifier": "Foo-0.2.3",
            "requirements": [
                "bim >= 3, < 4",
                "bar",
            ]
        },
        {
            "identifier": "Bar-22.3",
            "requirements": [
                "foo",
            ]
        },
        {
            "identifier": "Bim-3.2.1",
        }
    ]

    context = {
        "environ": {
            "PYTHONPATH": "/path/to/site-packages"
        }
    }

    mocked_tempfile_mkdtemp.side_effect = ["/tmp1", "/tmp2"]
    mocked_fetch_context_mapping.return_value = context
    mocked_package_install.side_effect = packages

    mocked_copy_to_destination.side_effect = [
        (True, overwrite), (True, overwrite), (True, overwrite)
    ]

    qip.install(
        ["foo", "bar"], "/path/to/install",
        definition_path="/path/to/definitions",
        **options
    )

    assert mocked_tempfile_mkdtemp.call_count == 2

    assert mocked_filesystem_ensure_directory.call_count == 8
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/install")
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/definitions")
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/site-packages")
    mocked_filesystem_ensure_directory.assert_any_call("/tmp2")

    mocked_fetch_context_mapping.assert_called_once_with("/tmp2", python_target)

    assert mocked_package_install.call_count == 3
    mocked_package_install.assert_any_call(
        "foo", "/tmp2", context, "/tmp1",
        editable_mode=editable_mode
    )
    mocked_package_install.assert_any_call(
        "bar", "/tmp2", context, "/tmp1",
        editable_mode=False
    )
    mocked_package_install.assert_any_call(
        "bim >= 3, < 4", "/tmp2", context, "/tmp1",
        editable_mode=False
    )

    assert mocked_copy_to_destination.call_count == 3
    mocked_copy_to_destination.assert_any_call(
        packages[0], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )
    mocked_copy_to_destination.assert_any_call(
        packages[1], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )
    mocked_copy_to_destination.assert_any_call(
        packages[2], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )

    assert mocked_definition_export.call_count == 3
    mocked_definition_export.assert_any_call(
        "/path/to/definitions", packages[0], "/path/to/install",
        editable_mode=editable_mode,
        definition_mapping=definition_mapping
    )

    mocked_definition_export.assert_any_call(
        "/path/to/definitions", packages[1], "/path/to/install",
        editable_mode=False,
        definition_mapping=definition_mapping
    )
    mocked_definition_export.assert_any_call(
        "/path/to/definitions", packages[2], "/path/to/install",
        editable_mode=False,
        definition_mapping=definition_mapping
    )

    assert mocked_shutil_rmtree.call_count == 5
    mocked_shutil_rmtree.assert_any_call("/tmp1")
    mocked_shutil_rmtree.assert_any_call("/tmp2")


@pytest.mark.parametrize(
    "options, overwrite, editable_mode, python_target", [
        ({}, False, False, sys.executable),
        ({"overwrite": True}, True, False, sys.executable),
        ({"editable_mode": True}, False, True, sys.executable),
        ({"python_target": "/bin/python3"}, False, False, "/bin/python3")
    ], ids=[
        "no-options",
        "with-overwrite-packages",
        "with-editable-mode",
        "with-python-target",
    ]
)
def test_install_without_dependencies(
    mocked_filesystem_ensure_directory, mocked_tempfile_mkdtemp,
    mocked_fetch_context_mapping, mocked_package_install,
    mocked_copy_to_destination, mocked_definition_export,
    mocked_shutil_rmtree, options, overwrite, editable_mode, python_target
):
    """Install packages with dependencies."""
    packages = [
        {
            "identifier": "Foo-0.2.3",
            "requirements": [
                "bim >= 3, < 4",
                "bar",
            ]
        },
        {
            "identifier": "Bar-22.3",
            "requirements": [
                "foo",
            ]
        }
    ]

    context = {
        "environ": {
            "PYTHONPATH": "/path/to/site-packages"
        }
    }

    mocked_tempfile_mkdtemp.side_effect = ["/tmp1", "/tmp2"]
    mocked_fetch_context_mapping.return_value = context
    mocked_package_install.side_effect = packages

    mocked_copy_to_destination.side_effect = [
        (True, overwrite), (True, overwrite)
    ]

    qip.install(
        ["foo", "bar"], "/path/to/install", no_dependencies=True, **options
    )

    assert mocked_tempfile_mkdtemp.call_count == 2

    assert mocked_filesystem_ensure_directory.call_count == 5
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/install")
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/site-packages")
    mocked_filesystem_ensure_directory.assert_any_call("/tmp2")

    mocked_fetch_context_mapping.assert_called_once_with("/tmp2", python_target)

    assert mocked_package_install.call_count == 2
    mocked_package_install.assert_any_call(
        "foo", "/tmp2", context, "/tmp1",
        editable_mode=editable_mode
    )
    mocked_package_install.assert_any_call(
        "bar", "/tmp2", context, "/tmp1",
        editable_mode=False
    )

    assert mocked_copy_to_destination.call_count == 2
    mocked_copy_to_destination.assert_any_call(
        packages[0], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )
    mocked_copy_to_destination.assert_any_call(
        packages[1], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )

    mocked_definition_export.assert_not_called()

    assert mocked_shutil_rmtree.call_count == 4
    mocked_shutil_rmtree.assert_any_call("/tmp1")
    mocked_shutil_rmtree.assert_any_call("/tmp2")


@pytest.mark.parametrize(
    "options, overwrite, editable_mode, python_target", [
        ({}, False, False, sys.executable),
        ({"overwrite": True}, True, False, sys.executable),
        ({"editable_mode": True}, False, True, sys.executable),
        ({"python_target": "/bin/python3"}, False, False, "/bin/python3")
    ], ids=[
        "no-options",
        "with-overwrite-packages",
        "with-editable-mode",
        "with-python-target",
    ]
)
def test_install_with_package_skipped(
    mocked_filesystem_ensure_directory, mocked_tempfile_mkdtemp,
    mocked_fetch_context_mapping, mocked_package_install,
    mocked_copy_to_destination, mocked_definition_export,
    mocked_shutil_rmtree, options, overwrite, editable_mode, python_target
):
    """Install packages with one package copy skipped."""
    packages = [
        {
            "identifier": "Foo-0.2.3",
            "requirements": [
                "bim >= 3, < 4",
                "bar",
            ]
        },
        {
            "identifier": "Bar-22.3",
            "requirements": [
                "foo",
            ]
        },
        {
            "identifier": "Bim-3.2.1",
        }
    ]

    context = {
        "environ": {
            "PYTHONPATH": "/path/to/site-packages"
        }
    }

    mocked_tempfile_mkdtemp.side_effect = ["/tmp1", "/tmp2"]
    mocked_fetch_context_mapping.return_value = context
    mocked_package_install.side_effect = packages

    mocked_copy_to_destination.return_value = (False, overwrite)

    qip.install(["foo"], "/path/to/install", **options)

    assert mocked_tempfile_mkdtemp.call_count == 2

    assert mocked_filesystem_ensure_directory.call_count == 7
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/install")
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/site-packages")
    mocked_filesystem_ensure_directory.assert_any_call("/tmp2")

    mocked_fetch_context_mapping.assert_called_once_with("/tmp2", python_target)

    assert mocked_package_install.call_count == 3
    mocked_package_install.assert_any_call(
        "foo", "/tmp2", context, "/tmp1",
        editable_mode=editable_mode
    )

    assert mocked_copy_to_destination.call_count == 3
    mocked_copy_to_destination.assert_any_call(
        packages[0], "/tmp2", "/path/to/install",
        overwrite=overwrite
    )

    mocked_definition_export.assert_not_called()

    assert mocked_shutil_rmtree.call_count == 5
    mocked_shutil_rmtree.assert_any_call("/tmp1")
    mocked_shutil_rmtree.assert_any_call("/tmp2")


def test_install_with_package_installation_error(
    mocked_filesystem_ensure_directory, mocked_tempfile_mkdtemp,
    mocked_fetch_context_mapping, mocked_package_install,
    mocked_copy_to_destination, mocked_definition_export,
    mocked_shutil_rmtree, logger
):
    """Install packages with one package error which is skipped."""
    package = {"identifier": "Foo-0.2.3"}

    context = {
        "environ": {
            "PYTHONPATH": "/path/to/site-packages"
        }
    }

    mocked_tempfile_mkdtemp.side_effect = ["/tmp1", "/tmp2"]
    mocked_fetch_context_mapping.return_value = context
    mocked_package_install.side_effect = [RuntimeError("Oops"), package]

    mocked_copy_to_destination.return_value = (True, True)

    qip.install(["foo", "bar"], "/path/to/install")

    assert mocked_tempfile_mkdtemp.call_count == 2

    assert mocked_filesystem_ensure_directory.call_count == 5
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/install")
    mocked_filesystem_ensure_directory.assert_any_call("/path/to/site-packages")
    mocked_filesystem_ensure_directory.assert_any_call("/tmp2")

    mocked_fetch_context_mapping.assert_called_once_with(
        "/tmp2", sys.executable
    )

    assert mocked_package_install.call_count == 2
    mocked_package_install.assert_any_call(
        "foo", "/tmp2", context, "/tmp1",
        editable_mode=False
    )
    mocked_package_install.assert_any_call(
        "bar", "/tmp2", context, "/tmp1",
        editable_mode=False
    )

    mocked_copy_to_destination.assert_called_once_with(
        package, "/tmp2", "/path/to/install", overwrite=False
    )

    mocked_definition_export.assert_not_called()

    assert mocked_shutil_rmtree.call_count == 4
    mocked_shutil_rmtree.assert_any_call("/tmp1")
    mocked_shutil_rmtree.assert_any_call("/tmp2")

    logger.error.assert_called_once_with("Request 'foo' has failed :\nOops")


def test_copy_to_destination(
    mocked_click_prompt, mocked_shutil_rmtree, mocked_shutil_copytree,
    mocked_filesystem_ensure_directory, logger
):
    """Copy package to destination."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "target": "Foo/Foo-0.2.3"
    }

    result = qip.copy_to_destination(
        mapping, "/path/to/installed/package", "/path/to/destination"
    )
    assert result == (True, False)

    mocked_click_prompt.assert_not_called()
    mocked_shutil_rmtree.assert_not_called()

    mocked_shutil_copytree.assert_called_once_with(
        "/path/to/installed/package",
        "/path/to/destination/Foo/Foo-0.2.3"
    )

    mocked_filesystem_ensure_directory.assert_called_once_with(
        "/path/to/destination/Foo"
    )

    logger.warning.assert_not_called()
    logger.info.assert_called_once_with("\tInstalled 'Foo-0.2.3'.")


def test_copy_to_destination_with_system_restriction(
    mocked_click_prompt, mocked_shutil_rmtree, mocked_shutil_copytree,
    mocked_filesystem_ensure_directory, logger
):
    """Copy package with system restriction to destination."""
    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "system": {
            "os": {
                "name": "centos",
                "major_version": 7
            }
        },
        "target": "Foo/Foo-0.2.3-centos7"
    }

    result = qip.copy_to_destination(
        mapping, "/path/to/installed/package", "/path/to/destination"
    )
    assert result == (True, False)

    mocked_click_prompt.assert_not_called()
    mocked_shutil_rmtree.assert_not_called()

    mocked_shutil_copytree.assert_called_once_with(
        "/path/to/installed/package",
        "/path/to/destination/Foo/Foo-0.2.3-centos7"
    )

    mocked_filesystem_ensure_directory.assert_called_once_with(
        "/path/to/destination/Foo"
    )

    logger.warning.assert_not_called()
    logger.info.assert_called_once_with("\tInstalled 'Foo-0.2.3'.")


def test_copy_to_destination_skip_existing(
    temporary_directory, mocked_click_prompt, mocked_shutil_rmtree,
    mocked_shutil_copytree, mocked_filesystem_ensure_directory, logger
):
    """Copy package to destination by skipping existing package."""
    path = os.path.join(temporary_directory, "Foo", "Foo-0.2.3")
    os.makedirs(path)

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "target": "Foo/Foo-0.2.3"
    }

    result = qip.copy_to_destination(
        mapping, "/path/to/installed/package", temporary_directory
    )
    assert result == (False, False)

    mocked_click_prompt.assert_not_called()
    mocked_shutil_rmtree.assert_not_called()

    mocked_shutil_copytree.assert_not_called()
    mocked_filesystem_ensure_directory.assert_not_called()

    logger.warning.assert_called_once_with(
        "Skip 'Foo-0.2.3' which is already installed."
    )
    logger.info.assert_not_called()


def test_copy_to_destination_overwrite_existing(
    temporary_directory, mocked_click_prompt, mocked_shutil_rmtree,
    mocked_shutil_copytree, mocked_filesystem_ensure_directory, logger
):
    """Copy package to destination by overwriting existing package."""
    path = os.path.join(temporary_directory, "Foo", "Foo-0.2.3")
    os.makedirs(path)

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "target": "Foo/Foo-0.2.3"
    }

    result = qip.copy_to_destination(
        mapping, "/path/to/installed/package", temporary_directory,
        overwrite=True
    )
    assert result == (True, True)

    mocked_click_prompt.assert_not_called()

    mocked_shutil_rmtree.assert_called_once_with(path)

    mocked_shutil_copytree.assert_called_once_with(
        "/path/to/installed/package", path
    )

    mocked_filesystem_ensure_directory.assert_called_once_with(
        os.path.join(temporary_directory, "Foo")
    )

    logger.warning.assert_called_once_with(
        "Overwrite 'Foo-0.2.3' which is already installed."
    )
    logger.info.assert_called_once_with("\tInstalled 'Foo-0.2.3'.")


@pytest.mark.parametrize("overwrite, overwrite_next, expected", [
    (True, None, None),
    (False, None, None),
    (True, True, True),
    (False, False, False),
], ids=[
    "yes",
    "no",
    "yes-to-all",
    "no-to-all",
])
def test_copy_to_destination_confirm_overwrite(
    temporary_directory, mocked_confirm_overwrite, mocked_shutil_rmtree,
    mocked_shutil_copytree, mocked_filesystem_ensure_directory, logger,
    overwrite, overwrite_next, expected
):
    """Ask user to confirm overwrite existing package."""
    path = os.path.join(temporary_directory, "Foo", "Foo-0.2.3")
    os.makedirs(path)

    mocked_confirm_overwrite.return_value = (overwrite, overwrite_next)

    mapping = {
        "identifier": "Foo-0.2.3",
        "name": "Foo",
        "target": "Foo/Foo-0.2.3"
    }

    result = qip.copy_to_destination(
        mapping, "/path/to/installed/package", temporary_directory,
        overwrite=None
    )
    assert result == (overwrite, expected)

    if overwrite:
        mocked_shutil_rmtree.assert_called_once()
        mocked_shutil_copytree.assert_called_once()
        mocked_filesystem_ensure_directory.assert_called_once()
        logger.warning.assert_called_once()

    else:
        mocked_shutil_rmtree.assert_not_called()
        mocked_shutil_copytree.assert_not_called()
        mocked_filesystem_ensure_directory.assert_not_called()

        logger.warning.assert_called_once_with(
            "Skip 'Foo-0.2.3' which is already installed."
        )
        logger.info.assert_not_called()


@pytest.mark.parametrize("answer, expected", [
    ("y", (True, None)),
    ("n", (False, None)),
    ("ya", (True, True)),
    ("na", (False, False)),
], ids=[
    "yes",
    "no",
    "yes-to-all",
    "no-to-all",
])
def test_confirm_overwrite(mocker, mocked_click_prompt, answer, expected):
    """Ask user to confirm overwrite existing package."""
    # User answered
    mocked_click_prompt.return_value = answer

    result = qip._confirm_overwrite("foo")

    assert result == expected
    mocked_click_prompt.assert_called_once_with(
        "Overwrite 'foo'? ([y]es, [n]o, [ya] yes to all, [na] no to all)",
        default='n',
        show_choices=False,
        show_default=False,
        type=mocker.ANY
    )


def test_fetch_context_mapping(
    mocked_fetch_environ, mocked_fetch_python_mapping
):
    """Return context mapping containing environment and python mapping."""
    mocked_fetch_environ.return_value = {
        "PATH": "/path/to/bin",
        "PYTHONPATH": "/path/to/lib",
    }
    mocked_fetch_python_mapping.return_value = {
        "library-path": "lib/python2.7/site-packages"
    }

    assert qip.fetch_context_mapping("/path", "python==2.7.*") == {
        "environ": {
            "PATH": "/path/to/bin",
            "PYTHONPATH": "/path/lib/python2.7/site-packages",
        },
        "python": {
            "library-path": "lib/python2.7/site-packages"
        }
    }



# :coding: utf-8

import os
import sys
import tempfile
import shutil

import six.moves
import click
import wiz.filesystem

import qip.definition
import qip.package
import qip.environ
import qip.logging

from qip._version import __version__


def install(
    requests, output_path, definition_path=None, overwrite=False,
    no_dependencies=False, editable_mode=False, python_target=sys.executable,
    definition_mapping=None
):
    """Install packages to *output_path* from *requests*.

    :param requests: list of packages to be installed

        A request can be one of::

            ["."]
            ["/path/to/foo/"]
            ["foo", "bar"]
            ["foo==0.1.0"]
            ["foo >= 7, < 8"]
            ["git@gitlab:rnd/foo.git"]
            ["git@gitlab:rnd/foo.git@0.1.0"]
            ["git@gitlab:rnd/foo.git@dev"]

    :param output_path: root destination path for Python packages installation.

    :param definition_path: :term:`Wiz` definition extraction path. Default is
        None, which means that :term:`Wiz` definitions are not extracted.

    :param overwrite: indicate whether packages already installed and
        corresponding :term:`Wiz` definitions should be overwritten. If None, a
        user confirmation will be prompted. Default is False.

    :param no_dependencies: indicate whether package dependencies should be
        skipped. Default is False.

    :param editable_mode: indicate whether the Python package location should
        target the source installation package. Default is False.

    :param python_target: Target a specific Python version via a Wiz request or
        a path to a Python executable (e.g. "python==2.7.*" or
        "/path/to/bin/python"). Default is the path to the current Python
        executable.

    :param definition_mapping: None or mapping regrouping all available
        :term:`Wiz` definitions. Default is None.

    :return: Boolean value.

    """
    logger = qip.logging.Logger(__name__ + ".install")

    wiz.filesystem.ensure_directory(output_path)
    if definition_path is not None:
        wiz.filesystem.ensure_directory(definition_path)

    # Setup temporary folder for package installation.
    cache_path = tempfile.mkdtemp()
    package_path = tempfile.mkdtemp()

    try:
        # Fetch environment mapping and installation path.
        context_mapping = fetch_context_mapping(package_path, python_target)
        library_path = context_mapping["environ"]["PYTHONPATH"]

        # Record requests and package installed to prevent duplications.
        installed_packages = set()
        installed_requests = set()

        # Fill up queue with requirements extracted from requests.
        queue = six.moves.queue.Queue()

        for request in requests:
            queue.put((request, None))

        while not queue.empty():
            request, parent_identifier = queue.get()
            if request in installed_requests:
                continue

            # Clean up before installation.
            logger.debug("Clean up directory content before installation")
            shutil.rmtree(package_path)
            wiz.filesystem.ensure_directory(package_path)

            # Needed for the editable mode.
            wiz.filesystem.ensure_directory(library_path)

            try:
                package_mapping = qip.package.install(
                    request, package_path, context_mapping, cache_path,
                    editable_mode=editable_mode
                )

            except RuntimeError as error:
                prompt = "Request '{}' has failed ".format(request)
                if parent_identifier is not None:
                    prompt += " [from '{}']".format(parent_identifier)

                logger.error("{}:\n{}".format(prompt, error))
                continue

            if package_mapping["identifier"] in installed_packages:
                continue

            prompt = "Requested '{}'".format(request)
            if parent_identifier is not None:
                prompt += " [from '{}'].".format(parent_identifier)
            logger.info(prompt)

            installed_packages.add(package_mapping["identifier"])
            installed_requests.add(request)

            # Install package to destination.
            success, overwrite = copy_to_destination(
                package_mapping,
                package_path,
                output_path,
                overwrite=overwrite
            )

            # Extract a wiz definition is requested.
            if success and definition_path is not None:
                qip.definition.export(
                    definition_path,
                    package_mapping,
                    output_path,
                    editable_mode=editable_mode,
                    definition_mapping=definition_mapping
                )

            # Reset editable mode to False for requirements.
            editable_mode = False

            # Fill up queue with requirements extracted from package
            # dependencies.
            if not no_dependencies:
                for request in package_mapping.get("requirements", []):
                    queue.put((request, package_mapping["identifier"]))

    finally:
        shutil.rmtree(package_path)
        shutil.rmtree(cache_path)

    logger.info(
        "Packages installed: {}".format(
            ", ".join(sorted(installed_packages, key=lambda s: s.lower()))
        )
    )


def copy_to_destination(
    mapping, source_path, destination_path, overwrite=False
):
    """Copy package from *source_path* to *destination_path*.

    :param mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param source_path: path where the package was built.

    :param destination_path: path to install to.

    :param overwrite: indicate whether packages already installed should be
        overwritten. If None, a user confirmation will be prompted. Default is
        False.

    :return: tuple with one boolean value indicating whether the copy has been
        done and one indicating a new value for the *overwrite* option.

    """
    logger = qip.logging.Logger(__name__ + ".copy_to_destination")

    identifier = mapping["identifier"]
    target = os.path.join(destination_path, mapping["target"])

    # By default, future overwrite request will the same as the present one.
    overwrite_next = overwrite

    if os.path.isdir(target):
        if overwrite is None:
            overwrite, overwrite_next = _confirm_overwrite(identifier)

        if overwrite:
            logger.warning(
                "Overwrite '{}' which is already installed.".format(identifier)
            )
            shutil.rmtree(target)

        else:
            logger.warning(
                "Skip '{}' which is already installed.".format(identifier)
            )

            return False, overwrite_next

    wiz.filesystem.ensure_directory(os.path.dirname(target))
    shutil.copytree(source_path, target)
    logger.debug("Source copied to '{}'".format(target))

    logger.info("\tInstalled '{}'.".format(identifier))

    return True, overwrite_next


def _confirm_overwrite(identifier):
    """Package overwrite confirmation prompt for *identifier*

    Return a tuple with one boolean value indicating whether *identifier* should
    be overwritten and one boolean indicating whether future packages should
    also be overwritten.

    """
    message = (
        "Overwrite '{}'? ([y]es, [n]o, [ya] yes to all, [na] no to all)"
        .format(identifier)
    )

    answer = click.prompt(
        message,
        type=click.Choice(["y", "n", "ya", "na"]),
        default="n",
        show_choices=False,
        show_default=False
    )

    overwrite = answer[0] == "y"
    overwrite_next = overwrite if "a" in answer else None
    return overwrite, overwrite_next


def fetch_context_mapping(path, python_target):
    """Return context mapping containing environment and python mapping.

    :param path: path where python package has been installed.

    :param python_target: Target a specific Python version via a Wiz request or
        a path to a Python executable (e.g. "python==2.7.*" or
        "/path/to/bin/python").

    :return: Context mapping.

        It should be in the form of::

            {
                "environ": {
                    "PATH": "/path/to/bin",
                    "PYTHONPATH": "/path/to/lib/python2.7/site-packages",
                },
                "python": {
                    "identifier": "2.7",
                    "request": "python >= 2.7, < 2.8",
                    "installation-target": "lib/python2.7/site-packages"
                }
            }

    """
    environ_mapping = qip.environ.fetch(
        python_target, mapping={"PYTHONWARNINGS": "ignore:DEPRECATION"}
    )

    # Fetch Python version mapping from environment
    python_mapping = qip.environ.fetch_python_mapping(environ_mapping)

    # Compute the installation path and add it to PYTHONPATH
    install_path = os.path.join(path, python_mapping["library-path"])
    environ_mapping["PYTHONPATH"] = install_path

    return {
        "environ": environ_mapping,
        "python": python_mapping
    }
